# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import unittest

from opentelemetry import baggage, trace
from opentelemetry.propagate import extract, inject
from opentelemetry.trace import get_current_span, set_span_in_context
from opentelemetry.trace.span import format_span_id, format_trace_id


class TestDefaultGlobalPropagator(unittest.TestCase):
    """Test ensures the default global composite propagator works as intended"""

    TRACE_ID = int("12345678901234567890123456789012", 16)  # type:int
    SPAN_ID = int("1234567890123456", 16)  # type:int

    def test_propagation(self):
        traceparent_value = (
            f"00-{format_trace_id(self.TRACE_ID)}-"
            f"{format_span_id(self.SPAN_ID)}-00"
        )
        tracestate_value = "foo=1,bar=2,baz=3"
        headers = {
            "baggage": ["key1=val1,key2=val2"],
            "traceparent": [traceparent_value],
            "tracestate": [tracestate_value],
        }
        ctx = extract(headers)
        baggage_entries = baggage.get_all(context=ctx)
        expected = {"key1": "val1", "key2": "val2"}
        self.assertEqual(baggage_entries, expected)
        span_context = get_current_span(context=ctx).get_span_context()

        self.assertEqual(span_context.trace_id, self.TRACE_ID)
        self.assertEqual(span_context.span_id, self.SPAN_ID)

        span = trace.NonRecordingSpan(span_context)
        ctx = baggage.set_baggage("key3", "val3")
        ctx = baggage.set_baggage("key4", "val4", context=ctx)
        ctx = set_span_in_context(span, context=ctx)
        output = {}
        inject(output, context=ctx)
        self.assertEqual(traceparent_value, output["traceparent"])
        self.assertIn("key3=val3", output["baggage"])
        self.assertIn("key4=val4", output["baggage"])
        self.assertIn("foo=1", output["tracestate"])
        self.assertIn("bar=2", output["tracestate"])
        self.assertIn("baz=3", output["tracestate"])



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import reload
from os import environ
from unittest import TestCase
from unittest.mock import Mock, patch

from opentelemetry.util import _providers


class Test_Providers(TestCase):  # pylint: disable=invalid-name
    @patch.dict(
        environ,
        {  # type: ignore
            "provider_environment_variable": "mock_provider_environment_variable"
        },
    )
    @patch("opentelemetry.util._importlib_metadata.entry_points")
    def test__providers(self, mock_entry_points):
        reload(_providers)

        mock_entry_points.configure_mock(
            **{
                "side_effect": [
                    [
                        Mock(
                            **{
                                "load.return_value": Mock(
                                    **{"return_value": "a"}
                                )
                            }
                        ),
                    ],
                ]
            }
        )

        self.assertEqual(
            _providers._load_provider(  # pylint: disable=protected-access
                "provider_environment_variable", "provider"
            ),
            "a",
        )



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import typing

from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.propagators import textmap
from opentelemetry.trace import format_span_id, format_trace_id
from opentelemetry.trace.span import TraceState


class TraceContextTextMapPropagator(textmap.TextMapPropagator):
    """Extracts and injects using w3c TraceContext's headers."""

    _TRACEPARENT_HEADER_NAME = "traceparent"
    _TRACESTATE_HEADER_NAME = "tracestate"
    _TRACEPARENT_HEADER_FORMAT = (
        "^[ \t]*([0-9a-f]{2})-([0-9a-f]{32})-([0-9a-f]{16})-([0-9a-f]{2})"
        + "(-.*)?[ \t]*$"
    )
    _TRACEPARENT_HEADER_FORMAT_RE = re.compile(_TRACEPARENT_HEADER_FORMAT)

    def extract(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        getter: textmap.Getter[textmap.CarrierT] = textmap.default_getter,
    ) -> Context:
        """Extracts SpanContext from the carrier.

        See `opentelemetry.propagators.textmap.TextMapPropagator.extract`
        """
        if context is None:
            context = Context()

        header = getter.get(carrier, self._TRACEPARENT_HEADER_NAME)

        if not header:
            return context

        match = re.search(self._TRACEPARENT_HEADER_FORMAT_RE, header[0])
        if not match:
            return context

        version: str = match.group(1)
        trace_id: str = match.group(2)
        span_id: str = match.group(3)
        trace_flags: str = match.group(4)

        if trace_id == "0" * 32 or span_id == "0" * 16:
            return context

        if version == "00":
            if match.group(5):  # type: ignore
                return context
        if version == "ff":
            return context

        tracestate_headers = getter.get(carrier, self._TRACESTATE_HEADER_NAME)
        if tracestate_headers is None:
            tracestate = None
        else:
            tracestate = TraceState.from_header(tracestate_headers)

        span_context = trace.SpanContext(
            trace_id=int(trace_id, 16),
            span_id=int(span_id, 16),
            is_remote=True,
            trace_flags=trace.TraceFlags(int(trace_flags, 16)),
            trace_state=tracestate,
        )
        return trace.set_span_in_context(
            trace.NonRecordingSpan(span_context), context
        )

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: typing.Optional[Context] = None,
        setter: textmap.Setter[textmap.CarrierT] = textmap.default_setter,
    ) -> None:
        """Injects SpanContext into the carrier.

        See `opentelemetry.propagators.textmap.TextMapPropagator.inject`
        """
        span = trace.get_current_span(context)
        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return
        traceparent_string = f"00-{format_trace_id(span_context.trace_id)}-{format_span_id(span_context.span_id)}-{span_context.trace_flags:02x}"
        setter.set(carrier, self._TRACEPARENT_HEADER_NAME, traceparent_string)
        if span_context.trace_state:
            tracestate_string = span_context.trace_state.to_header()
            setter.set(
                carrier, self._TRACESTATE_HEADER_NAME, tracestate_string
            )

    @property
    def fields(self) -> typing.Set[str]:
        """Returns a set with the fields set in `inject`.

        See
        `opentelemetry.propagators.textmap.TextMapPropagator.fields`
        """
        return {self._TRACEPARENT_HEADER_NAME, self._TRACESTATE_HEADER_NAME}



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore

# pylint: disable=protected-access

from unittest import TestCase
from unittest.mock import Mock, patch

from pytest import fixture

import opentelemetry.metrics._internal as metrics_internal
from opentelemetry import metrics
from opentelemetry.environment_variables import OTEL_PYTHON_METER_PROVIDER
from opentelemetry.metrics import (
    NoOpMeter,
    NoOpMeterProvider,
    get_meter_provider,
    set_meter_provider,
)
from opentelemetry.metrics._internal import _ProxyMeter, _ProxyMeterProvider
from opentelemetry.metrics._internal.instrument import (
    _ProxyCounter,
    _ProxyGauge,
    _ProxyHistogram,
    _ProxyObservableCounter,
    _ProxyObservableGauge,
    _ProxyObservableUpDownCounter,
    _ProxyUpDownCounter,
)
from opentelemetry.test.globals_test import (
    MetricsGlobalsTest,
    reset_metrics_globals,
)

# FIXME Test that the instrument methods can be called concurrently safely.


@fixture
def reset_meter_provider():
    print(f"calling reset_metrics_globals() {reset_metrics_globals}")
    reset_metrics_globals()
    yield
    print("teardown - calling reset_metrics_globals()")
    reset_metrics_globals()


# pylint: disable=redefined-outer-name
def test_set_meter_provider(reset_meter_provider):
    """
    Test that the API provides a way to set a global default MeterProvider
    """

    mock = Mock()

    assert metrics_internal._METER_PROVIDER is None

    set_meter_provider(mock)

    assert metrics_internal._METER_PROVIDER is mock
    assert get_meter_provider() is mock


def test_set_meter_provider_calls_proxy_provider(reset_meter_provider):
    with patch(
        "opentelemetry.metrics._internal._PROXY_METER_PROVIDER"
    ) as mock_proxy_mp:
        assert metrics_internal._PROXY_METER_PROVIDER is mock_proxy_mp
        mock_real_mp = Mock()
        set_meter_provider(mock_real_mp)
        mock_proxy_mp.on_set_meter_provider.assert_called_once_with(
            mock_real_mp
        )


def test_get_meter_provider(reset_meter_provider):
    """
    Test that the API provides a way to get a global default MeterProvider
    """

    assert metrics_internal._METER_PROVIDER is None

    assert isinstance(get_meter_provider(), _ProxyMeterProvider)

    metrics._METER_PROVIDER = None

    with patch.dict(
        "os.environ", {OTEL_PYTHON_METER_PROVIDER: "test_meter_provider"}
    ):
        with patch("opentelemetry.metrics._internal._load_provider", Mock()):
            with patch(
                "opentelemetry.metrics._internal.cast",
                Mock(**{"return_value": "test_meter_provider"}),
            ):
                assert get_meter_provider() == "test_meter_provider"


class TestGetMeter(TestCase):
    def test_get_meter_parameters(self):
        """
        Test that get_meter accepts name, version and schema_url
        """
        try:
            NoOpMeterProvider().get_meter(
                "name", version="version", schema_url="schema_url"
            )
        except Exception as error:  # pylint: disable=broad-exception-caught
            self.fail(f"Unexpected exception raised: {error}")

    def test_invalid_name(self):
        """
        Test that when an invalid name is specified a working meter
        implementation is returned as a fallback.

        Test that the fallback meter name property keeps its original invalid
        value.

        Test that a message is logged reporting the specified value for the
        fallback meter is invalid.
        """
        meter = NoOpMeterProvider().get_meter("")

        self.assertTrue(isinstance(meter, NoOpMeter))

        self.assertEqual(meter.name, "")

        meter = NoOpMeterProvider().get_meter(None)

        self.assertTrue(isinstance(meter, NoOpMeter))

        self.assertEqual(meter.name, None)


class TestProxy(MetricsGlobalsTest, TestCase):
    def test_global_proxy_meter_provider(self):
        # Global get_meter_provider() should initially be a _ProxyMeterProvider
        # singleton

        proxy_meter_provider: _ProxyMeterProvider = get_meter_provider()
        self.assertIsInstance(proxy_meter_provider, _ProxyMeterProvider)
        self.assertIs(get_meter_provider(), proxy_meter_provider)

    def test_proxy_provider(self):
        proxy_meter_provider = _ProxyMeterProvider()

        # Should return a proxy meter when no real MeterProvider is set
        name = "foo"
        version = "1.2"
        schema_url = "schema_url"
        proxy_meter: _ProxyMeter = proxy_meter_provider.get_meter(
            name, version=version, schema_url=schema_url
        )
        self.assertIsInstance(proxy_meter, _ProxyMeter)

        # After setting a real meter provider on the proxy, it should notify
        # it's _ProxyMeters which should create their own real Meters
        mock_real_mp = Mock()
        proxy_meter_provider.on_set_meter_provider(mock_real_mp)
        mock_real_mp.get_meter.assert_called_once_with(
            name, version, schema_url
        )

        # After setting a real meter provider on the proxy, it should now return
        # new meters directly from the set real meter
        another_name = "bar"
        meter2 = proxy_meter_provider.get_meter(another_name)
        self.assertIsInstance(meter2, Mock)
        mock_real_mp.get_meter.assert_called_with(another_name, None, None)

    # pylint: disable=too-many-locals,too-many-statements
    def test_proxy_meter(self):
        meter_name = "foo"
        proxy_meter: _ProxyMeter = _ProxyMeterProvider().get_meter(meter_name)
        self.assertIsInstance(proxy_meter, _ProxyMeter)

        # Should be able to create proxy instruments
        name = "foo"
        unit = "s"
        description = "Foobar"
        callback = Mock()
        proxy_counter = proxy_meter.create_counter(
            name, unit=unit, description=description
        )
        proxy_updowncounter = proxy_meter.create_up_down_counter(
            name, unit=unit, description=description
        )
        proxy_histogram = proxy_meter.create_histogram(
            name, unit=unit, description=description
        )

        proxy_gauge = proxy_meter.create_gauge(
            name, unit=unit, description=description
        )

        proxy_observable_counter = proxy_meter.create_observable_counter(
            name, callbacks=[callback], unit=unit, description=description
        )
        proxy_observable_updowncounter = (
            proxy_meter.create_observable_up_down_counter(
                name, callbacks=[callback], unit=unit, description=description
            )
        )
        proxy_overvable_gauge = proxy_meter.create_observable_gauge(
            name, callbacks=[callback], unit=unit, description=description
        )
        self.assertIsInstance(proxy_counter, _ProxyCounter)
        self.assertIsInstance(proxy_updowncounter, _ProxyUpDownCounter)
        self.assertIsInstance(proxy_histogram, _ProxyHistogram)
        self.assertIsInstance(proxy_gauge, _ProxyGauge)
        self.assertIsInstance(
            proxy_observable_counter, _ProxyObservableCounter
        )
        self.assertIsInstance(
            proxy_observable_updowncounter, _ProxyObservableUpDownCounter
        )
        self.assertIsInstance(proxy_overvable_gauge, _ProxyObservableGauge)

        # Synchronous proxy instruments should be usable
        amount = 12
        attributes = {"foo": "bar"}
        proxy_counter.add(amount, attributes=attributes)
        proxy_updowncounter.add(amount, attributes=attributes)
        proxy_histogram.record(amount, attributes=attributes)
        proxy_gauge.set(amount, attributes=attributes)

        # Calling _ProxyMeterProvider.on_set_meter_provider() should cascade down
        # to the _ProxyInstruments which should create their own real instruments
        # from the real Meter to back their calls
        real_meter_provider = Mock()
        proxy_meter.on_set_meter_provider(real_meter_provider)
        real_meter_provider.get_meter.assert_called_once_with(
            meter_name, None, None
        )

        real_meter: Mock = real_meter_provider.get_meter()
        real_meter.create_counter.assert_called_once_with(
            name, unit, description
        )
        real_meter.create_up_down_counter.assert_called_once_with(
            name, unit, description
        )
        real_meter.create_histogram.assert_called_once_with(
            name, unit, description
        )
        real_meter.create_gauge.assert_called_once_with(
            name, unit, description
        )
        real_meter.create_observable_counter.assert_called_once_with(
            name, [callback], unit, description
        )
        real_meter.create_observable_up_down_counter.assert_called_once_with(
            name, [callback], unit, description
        )
        real_meter.create_observable_gauge.assert_called_once_with(
            name, [callback], unit, description
        )

        # The synchronous instrument measurement methods should call through to
        # the real instruments
        real_counter: Mock = real_meter.create_counter()
        real_updowncounter: Mock = real_meter.create_up_down_counter()
        real_histogram: Mock = real_meter.create_histogram()
        real_gauge: Mock = real_meter.create_gauge()
        real_counter.assert_not_called()
        real_updowncounter.assert_not_called()
        real_histogram.assert_not_called()
        real_gauge.assert_not_called()

        proxy_counter.add(amount, attributes=attributes)
        real_counter.add.assert_called_once_with(amount, attributes, None)
        proxy_updowncounter.add(amount, attributes=attributes)
        real_updowncounter.add.assert_called_once_with(
            amount, attributes, None
        )
        proxy_histogram.record(amount, attributes=attributes)
        real_histogram.record.assert_called_once_with(amount, attributes, None)
        proxy_gauge.set(amount, attributes=attributes)
        real_gauge.set.assert_called_once_with(amount, attributes, None)

    def test_proxy_meter_with_real_meter(self) -> None:
        # Creating new instruments on the _ProxyMeter with a real meter set
        # should create real instruments instead of proxies
        meter_name = "foo"
        proxy_meter: _ProxyMeter = _ProxyMeterProvider().get_meter(meter_name)
        self.assertIsInstance(proxy_meter, _ProxyMeter)

        real_meter_provider = Mock()
        proxy_meter.on_set_meter_provider(real_meter_provider)

        name = "foo"
        unit = "s"
        description = "Foobar"
        callback = Mock()
        counter = proxy_meter.create_counter(
            name, unit=unit, description=description
        )
        updowncounter = proxy_meter.create_up_down_counter(
            name, unit=unit, description=description
        )
        histogram = proxy_meter.create_histogram(
            name, unit=unit, description=description
        )
        gauge = proxy_meter.create_gauge(
            name, unit=unit, description=description
        )
        observable_counter = proxy_meter.create_observable_counter(
            name, callbacks=[callback], unit=unit, description=description
        )
        observable_updowncounter = (
            proxy_meter.create_observable_up_down_counter(
                name, callbacks=[callback], unit=unit, description=description
            )
        )
        observable_gauge = proxy_meter.create_observable_gauge(
            name, callbacks=[callback], unit=unit, description=description
        )

        real_meter: Mock = real_meter_provider.get_meter()
        self.assertIs(counter, real_meter.create_counter())
        self.assertIs(updowncounter, real_meter.create_up_down_counter())
        self.assertIs(histogram, real_meter.create_histogram())
        self.assertIs(gauge, real_meter.create_gauge())
        self.assertIs(
            observable_counter, real_meter.create_observable_counter()
        )
        self.assertIs(
            observable_updowncounter,
            real_meter.create_observable_up_down_counter(),
        )
        self.assertIs(observable_gauge, real_meter.create_observable_gauge())



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import unittest
from typing import MutableSequence

from opentelemetry.attributes import BoundedAttributes, _clean_attribute


class TestAttributes(unittest.TestCase):
    # pylint: disable=invalid-name
    def assertValid(self, value, key="k"):
        expected = value
        if isinstance(value, MutableSequence):
            expected = tuple(value)
        self.assertEqual(_clean_attribute(key, value, None), expected)

    def assertInvalid(self, value, key="k"):
        self.assertIsNone(_clean_attribute(key, value, None))

    def test_attribute_key_validation(self):
        # only non-empty strings are valid keys
        self.assertInvalid(1, "")
        self.assertInvalid(1, 1)
        self.assertInvalid(1, {})
        self.assertInvalid(1, [])
        self.assertInvalid(1, b"1")
        self.assertValid(1, "k")
        self.assertValid(1, "1")

    def test_clean_attribute(self):
        self.assertInvalid([1, 2, 3.4, "ss", 4])
        self.assertInvalid([{}, 1, 2, 3.4, 4])
        self.assertInvalid(["sw", "lf", 3.4, "ss"])
        self.assertInvalid([1, 2, 3.4, 5])
        self.assertInvalid({})
        self.assertInvalid([1, True])
        self.assertValid(True)
        self.assertValid("hi")
        self.assertValid(3.4)
        self.assertValid(15)
        self.assertValid([1, 2, 3, 5])
        self.assertValid([1.2, 2.3, 3.4, 4.5])
        self.assertValid([True, False])
        self.assertValid(["ss", "dw", "fw"])
        self.assertValid([])
        # None in sequences are valid
        self.assertValid(["A", None, None])
        self.assertValid(["A", None, None, "B"])
        self.assertValid([None, None])
        self.assertInvalid(["A", None, 1])
        self.assertInvalid([None, "A", None, 1])

        # test keys
        self.assertValid("value", "key")
        self.assertInvalid("value", "")
        self.assertInvalid("value", None)

    def test_sequence_attr_decode(self):
        seq = [
            None,
            b"Content-Disposition",
            b"Content-Type",
            b"\x81",
            b"Keep-Alive",
        ]
        expected = [
            None,
            "Content-Disposition",
            "Content-Type",
            None,
            "Keep-Alive",
        ]
        self.assertEqual(
            _clean_attribute("headers", seq, None), tuple(expected)
        )


class TestBoundedAttributes(unittest.TestCase):
    # pylint: disable=consider-using-dict-items
    base = {
        "name": "Firulais",
        "age": 7,
        "weight": 13,
        "vaccinated": True,
    }

    def test_negative_maxlen(self):
        with self.assertRaises(ValueError):
            BoundedAttributes(-1)

    def test_from_map(self):
        dic_len = len(self.base)
        base_copy = self.base.copy()
        bdict = BoundedAttributes(dic_len, base_copy)

        self.assertEqual(len(bdict), dic_len)

        # modify base_copy and test that bdict is not changed
        base_copy["name"] = "Bruno"
        base_copy["age"] = 3

        for key in self.base:
            self.assertEqual(bdict[key], self.base[key])

        # test that iter yields the correct number of elements
        self.assertEqual(len(tuple(bdict)), dic_len)

        # map too big
        half_len = dic_len // 2
        bdict = BoundedAttributes(half_len, self.base)
        self.assertEqual(len(tuple(bdict)), half_len)
        self.assertEqual(bdict.dropped, dic_len - half_len)

    def test_bounded_dict(self):
        # create empty dict
        dic_len = len(self.base)
        bdict = BoundedAttributes(dic_len, immutable=False)
        self.assertEqual(len(bdict), 0)

        # fill dict
        for key in self.base:
            bdict[key] = self.base[key]

        self.assertEqual(len(bdict), dic_len)
        self.assertEqual(bdict.dropped, 0)

        for key in self.base:
            self.assertEqual(bdict[key], self.base[key])

        # test __iter__ in BoundedAttributes
        for key in bdict:
            self.assertEqual(bdict[key], self.base[key])

        # updating an existing element should not drop
        bdict["name"] = "Bruno"
        self.assertEqual(bdict.dropped, 0)

        # try to append more elements
        for key in self.base:
            bdict["new-" + key] = self.base[key]

        self.assertEqual(len(bdict), dic_len)
        self.assertEqual(bdict.dropped, dic_len)
        # Invalid values shouldn't be considered for `dropped`
        bdict["invalid-seq"] = [None, 1, "2"]
        self.assertEqual(bdict.dropped, dic_len)

        # test that elements in the dict are the new ones
        for key in self.base:
            self.assertEqual(bdict["new-" + key], self.base[key])

        # delete an element
        del bdict["new-name"]
        self.assertEqual(len(bdict), dic_len - 1)

        with self.assertRaises(KeyError):
            _ = bdict["new-name"]

    def test_no_limit_code(self):
        bdict = BoundedAttributes(maxlen=None, immutable=False)
        for num in range(100):
            bdict[str(num)] = num

        for num in range(100):
            self.assertEqual(bdict[str(num)], num)

    def test_immutable(self):
        bdict = BoundedAttributes()
        with self.assertRaises(TypeError):
            bdict["should-not-work"] = "dict immutable"

    def test_locking(self):
        """Supporting test case for a commit titled: Fix class BoundedAttributes to have RLock rather than Lock. See #3858.
        The change was introduced because __iter__ of the class BoundedAttributes holds lock, and we observed some deadlock symptoms
        in the codebase. This test case is to verify that the fix works as expected.
        """
        bdict = BoundedAttributes(immutable=False)

        with bdict._lock:  # pylint: disable=protected-access
            for num in range(100):
                bdict[str(num)] = num

        for num in range(100):
            self.assertEqual(bdict[str(num)], num)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from opentelemetry import trace
from opentelemetry.trace import TraceFlags, TraceState


class TestImmutableSpanContext(unittest.TestCase):
    def test_ctor(self):
        context = trace.SpanContext(
            1,
            1,
            is_remote=False,
            trace_flags=trace.DEFAULT_TRACE_OPTIONS,
            trace_state=trace.DEFAULT_TRACE_STATE,
        )

        self.assertEqual(context.trace_id, 1)
        self.assertEqual(context.span_id, 1)
        self.assertEqual(context.is_remote, False)
        self.assertEqual(context.trace_flags, trace.DEFAULT_TRACE_OPTIONS)
        self.assertEqual(context.trace_state, trace.DEFAULT_TRACE_STATE)

    def test_attempt_change_attributes(self):
        context = trace.SpanContext(
            1,
            2,
            is_remote=False,
            trace_flags=trace.DEFAULT_TRACE_OPTIONS,
            trace_state=trace.DEFAULT_TRACE_STATE,
        )

        # attempt to change the attribute values
        context.trace_id = 2  # type: ignore
        context.span_id = 3  # type: ignore
        context.is_remote = True  # type: ignore
        context.trace_flags = TraceFlags(3)  # type: ignore
        context.trace_state = TraceState([("test", "test")])  # type: ignore

        # check if attributes changed
        self.assertEqual(context.trace_id, 1)
        self.assertEqual(context.span_id, 2)
        self.assertEqual(context.is_remote, False)
        self.assertEqual(context.trace_flags, trace.DEFAULT_TRACE_OPTIONS)
        self.assertEqual(context.trace_state, trace.DEFAULT_TRACE_STATE)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# FIXME: Use importlib.metadata when support for 3.11 is dropped if the rest of
# the supported versions at that time have the same API.
from importlib_metadata import (  # type: ignore
    Distribution,
    EntryPoint,
    EntryPoints,
    PackageNotFoundError,
    distributions,
    entry_points,
    requires,
    version,
)

__all__ = [
    "entry_points",
    "version",
    "EntryPoint",
    "EntryPoints",
    "requires",
    "Distribution",
    "distributions",
    "PackageNotFoundError",
]



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import patch

from opentelemetry import context
from opentelemetry.context.context import Context
from opentelemetry.context.contextvars_context import ContextVarsRuntimeContext
from opentelemetry.environment_variables import OTEL_PYTHON_CONTEXT


def _do_work() -> str:
    key = context.create_key("say")
    context.attach(context.set_value(key, "bar"))
    return key


class TestContext(unittest.TestCase):
    def setUp(self):
        context.attach(Context())

    def test_context_key(self):
        key1 = context.create_key("say")
        key2 = context.create_key("say")
        self.assertNotEqual(key1, key2)
        first = context.set_value(key1, "foo")
        second = context.set_value(key2, "bar")
        self.assertEqual(context.get_value(key1, context=first), "foo")
        self.assertEqual(context.get_value(key2, context=second), "bar")

    def test_context(self):
        key1 = context.create_key("say")
        self.assertIsNone(context.get_value(key1))
        empty = context.get_current()
        second = context.set_value(key1, "foo")
        self.assertEqual(context.get_value(key1, context=second), "foo")

        key2 = _do_work()
        self.assertEqual(context.get_value(key2), "bar")
        third = context.get_current()

        self.assertIsNone(context.get_value(key1, context=empty))
        self.assertEqual(context.get_value(key1, context=second), "foo")
        self.assertEqual(context.get_value(key2, context=third), "bar")

    def test_set_value(self):
        first = context.set_value("a", "yyy")
        second = context.set_value("a", "zzz")
        third = context.set_value("a", "---", first)
        self.assertEqual("yyy", context.get_value("a", context=first))
        self.assertEqual("zzz", context.get_value("a", context=second))
        self.assertEqual("---", context.get_value("a", context=third))
        self.assertEqual(None, context.get_value("a"))

    def test_context_is_immutable(self):
        with self.assertRaises(ValueError):
            # ensure a context
            context.get_current()["test"] = "cant-change-immutable"

    def test_set_current(self):
        context.attach(context.set_value("a", "yyy"))

        token = context.attach(context.set_value("a", "zzz"))
        self.assertEqual("zzz", context.get_value("a"))

        context.detach(token)
        self.assertEqual("yyy", context.get_value("a"))


class TestInitContext(unittest.TestCase):
    def test_load_runtime_context_default(self):
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)

    @patch.dict("os.environ", {OTEL_PYTHON_CONTEXT: "contextvars_context"})
    def test_load_runtime_context(self):  # type: ignore[misc]
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)

    @patch.dict("os.environ", {OTEL_PYTHON_CONTEXT: "foo"})
    def test_load_runtime_context_fallback(self):  # type: ignore[misc]
        ctx = context._load_runtime_context()  # pylint: disable=W0212
        self.assertIsInstance(ctx, ContextVarsRuntimeContext)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from unittest import TestCase

from opentelemetry.trace import (
    INVALID_SPAN,
    NoOpTracer,
    Span,
    Tracer,
    _agnosticcontextmanager,
    get_current_span,
)


class TestTracer(TestCase):
    def setUp(self):
        self.tracer = NoOpTracer()

    def test_start_span(self):
        with self.tracer.start_span("") as span:
            self.assertIsInstance(span, Span)

    def test_start_as_current_span_context_manager(self):
        with self.tracer.start_as_current_span("") as span:
            self.assertIsInstance(span, Span)

    def test_start_as_current_span_decorator(self):
        # using a list to track the mock call order
        calls = []

        class MockTracer(Tracer):
            def start_span(self, *args, **kwargs):
                return INVALID_SPAN

            @_agnosticcontextmanager  # pylint: disable=protected-access
            def start_as_current_span(self, *args, **kwargs):  # type: ignore
                calls.append(1)
                yield INVALID_SPAN
                calls.append(9)

        mock_tracer = MockTracer()

        # test 1 : sync function
        @mock_tracer.start_as_current_span("name")
        def function_sync(data: str) -> int:
            calls.append(5)
            return len(data)

        calls = []
        res = function_sync("123")
        self.assertEqual(res, 3)
        self.assertEqual(calls, [1, 5, 9])

        # test 2 : async function
        @mock_tracer.start_as_current_span("name")
        async def function_async(data: str) -> int:
            calls.append(5)
            return len(data)

        calls = []
        res = asyncio.run(function_async("123"))
        self.assertEqual(res, 3)
        self.assertEqual(calls, [1, 5, 9])

    def test_get_current_span(self):
        with self.tracer.start_as_current_span("test") as span:
            get_current_span().set_attribute("test", "test")
            self.assertEqual(span, INVALID_SPAN)
            self.assertFalse(hasattr("span", "attributes"))



# type:ignore
import unittest
from unittest.mock import Mock, patch

import opentelemetry._events as events
from opentelemetry._events import (
    get_event_logger_provider,
    set_event_logger_provider,
)
from opentelemetry.test.globals_test import EventsGlobalsTest


class TestGlobals(EventsGlobalsTest, unittest.TestCase):
    def test_set_event_logger_provider(self):
        elp_mock = Mock()
        # pylint: disable=protected-access
        self.assertIsNone(events._EVENT_LOGGER_PROVIDER)
        set_event_logger_provider(elp_mock)
        self.assertIs(events._EVENT_LOGGER_PROVIDER, elp_mock)
        self.assertIs(get_event_logger_provider(), elp_mock)

    def test_get_event_logger_provider(self):
        # pylint: disable=protected-access
        self.assertIsNone(events._EVENT_LOGGER_PROVIDER)

        self.assertIsInstance(
            get_event_logger_provider(), events.ProxyEventLoggerProvider
        )

        events._EVENT_LOGGER_PROVIDER = None

        with patch.dict(
            "os.environ",
            {
                "OTEL_PYTHON_EVENT_LOGGER_PROVIDER": "test_event_logger_provider"
            },
        ):
            with patch("opentelemetry._events._load_provider", Mock()):
                with patch(
                    "opentelemetry._events.cast",
                    Mock(**{"return_value": "test_event_logger_provider"}),
                ):
                    self.assertEqual(
                        get_event_logger_provider(),
                        "test_event_logger_provider",
                    )
